from __future__ import annotations

from outrank.feature_transformations.feature_transformer_vault.default_transformers import DEFAULT_TRANSFORMERS
from outrank.feature_transformations.feature_transformer_vault.default_transformers import EXTENDED_ROUNDED_TRANSFORMERS
from outrank.feature_transformations.feature_transformer_vault.default_transformers import EXTENDED_TRANSFORMERS
from outrank.feature_transformations.feature_transformer_vault.default_transformers import MINIMAL_TRANSFORMERS
from outrank.feature_transformations.feature_transformer_vault.default_transformers import VERBOSE_TRANSFORMERS
from outrank.feature_transformations.feature_transformer_vault.fw_transformers import \
    FW_TRANSFORMERS

_tr_global_namespace = {
    'default': DEFAULT_TRANSFORMERS,
    'minimal': MINIMAL_TRANSFORMERS,
    'fw-transformers': FW_TRANSFORMERS,
    'extended': EXTENDED_TRANSFORMERS,
    'verbose': VERBOSE_TRANSFORMERS,
    'extended_rounded': EXTENDED_ROUNDED_TRANSFORMERS,
}



from __future__ import annotations

import numpy as np
import numpy.typing as npt

np.random.seed(123)
max_size = 10**6


def max_pair_coverage(array1: npt.NDArray[np.int32], array2: npt.NDArray[np.int32]) -> float:
    def hash_pair(el1: np.int32, el2: np.int32):
        return (el1 * 1471343 - el2) % max_size

    counts = np.zeros(max_size, dtype=np.int32)
    tot_len = len(array1)
    for i in range(tot_len):
        identifier = hash_pair(array1[i], array2[i])
        counts[identifier] += 1

    return np.max(counts) / tot_len


if __name__ == '__main__':

    array1 = np.array([1,1,2,3,1,1,1,5] * 100000)
    array2 = np.array([0,0,5,5,3,0,0,0] * 100000)
    coverage = max_pair_coverage(array1, array2)
    assert coverage == 0.5



from __future__ import annotations

import logging
import os
from collections import defaultdict

import numpy as np
import pandas as pd

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)


def outrank_task_result_summary(args):
    triplets_path = os.path.join(args.output_folder, 'pairwise_ranks.tsv')
    triplets = pd.read_csv(triplets_path, sep='\t')
    triplets = triplets.sort_values(by='Score', ascending=False)

    final_ranking = []
    for _, row in triplets.iterrows():
        feature_a, feature_b = row['FeatureA'], row['FeatureB']
        score = row['Score']
        if args.label_column == feature_a.split('-')[0]:
            final_ranking.append([feature_b, score])
        elif args.label_column == feature_b.split('-')[0]:
            final_ranking.append([feature_a, score])

    final_df = pd.DataFrame(final_ranking, columns=['Feature', f'Score {args.heuristic}'])
    final_df = (
        final_df.groupby('Feature')
        .median()
        .reset_index()
        .sort_values(by=f'Score {args.heuristic}', ascending=False)
    )

    if "MI" in args.heuristic:
        min_score = final_df[f'Score {args.heuristic}'].min()
        max_score = final_df[f'Score {args.heuristic}'].max()
        final_df[f'Score {args.heuristic}'] = (final_df[f'Score {args.heuristic}'] - min_score) / (max_score - min_score)

    logging.info(f'Storing summary files to {args.output_folder}')
    pd.set_option('display.max_rows', None, 'display.max_columns', None)

    singles_path = os.path.join(args.output_folder, 'feature_singles.tsv')
    final_df.to_csv(singles_path, sep='\t', index=False)

    if args.interaction_order > 1:
        feature_store = defaultdict(list)
        for _, row in final_df.iterrows():
            fname = row['Feature']
            score = row[f'Score {args.heuristic}']
            if 'AND' in fname:
                for el in fname.split('-')[0].split(' AND '):
                    feature_store[el].append(score)

        final_aggregate_df = pd.DataFrame([
            {
                'Feature': k,
                f'Combined score (order: {args.interaction_order}, {args.heuristic})': np.median(v),
            }
            for k, v in feature_store.items()
        ])
        final_aggregate_df.to_csv(
            os.path.join(args.output_folder, 'feature_singles_aggregated.tsv'), sep='\t', index=False
        )

    transformers_only_path = singles_path.replace('.tsv', '_transformers_only_imp.tsv')
    final_df[final_df['Feature'].str.contains('_tr_')].to_csv(transformers_only_path, sep='\t', index=False)



from __future__ import annotations

import logging
import os
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.cluster import hierarchy
from sklearn.manifold import TSNE
from sklearn.metrics import silhouette_score

from outrank.core_utils import read_reference_json

logging.basicConfig(format='%(asctime)s %(message)s', level=logging.INFO)
plt.rcParams['figure.figsize'] = (50, 30)


def visualize_hierarchical_clusters(
    triplet_dataframe: pd.DataFrame,
    output_folder: str,
    image_format: str = 'png',
    max_num_clusters: int = 100,
) -> None:
    plt.rcParams['figure.figsize'] = (10, 5)
    unique_features = triplet_dataframe.FeatureA.unique()

    if len(unique_features) > 1000:
        logging.info('Trying to visualize too many features, exiting ..')
        exit()

    dmat = np.zeros((len(unique_features), len(unique_features)))
    logging.info('Preparing the data for clustering ..')

    if triplet_dataframe.shape[0] > 10**5:
        logging.info('Trying to visualize more than 10 ** 5 triplets, exiting ..')
        exit()

    pivot_table = pd.pivot_table(
        triplet_dataframe,
        values='Score',
        index='FeatureA',
        columns='FeatureB',
        aggfunc=np.mean,
    )

    pivot_table.fillna(0, inplace=True)
    dmat = 1 - pivot_table.values

    logging.info('Clustering ..')

    for linkage_heuristic in ['complete']:
        Z = hierarchy.linkage(dmat, linkage_heuristic)

        hierarchy.dendrogram(
            Z, above_threshold_color='y', orientation='top', labels=unique_features,
        )
        plt.title(f'Linkage function: {linkage_heuristic}')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            plt.tight_layout()
        out_path = f'{output_folder}/dendrogram_{linkage_heuristic}.{image_format}'
        plt.savefig(out_path, dpi=300)
        plt.clf()
        plt.cla()
        logging.info(f'Visualized hierarchical clustering with linkage {linkage_heuristic} to {out_path}')

        range_min, range_max = np.min(pivot_table.values), np.max(pivot_table.values)
        spectrum = np.arange(range_min, range_max, (range_max - range_min) / 1000)
        max_silhouette = 0
        top_clustering = []
        full_silhouette_space = []

        for possible_threshold in spectrum:
            cluster_assignments = hierarchy.fcluster(Z, possible_threshold)
            num_clusters = len(np.unique(cluster_assignments))
            if num_clusters > 2 and num_clusters < max_num_clusters:
                try:
                    sil_score = silhouette_score(pivot_table, cluster_assignments)
                except Exception:
                    continue

                full_silhouette_space.append([sil_score, possible_threshold, num_clusters])
                if sil_score >= max_silhouette:
                    top_clustering = cluster_assignments
                    max_silhouette = sil_score

        dfx = pd.DataFrame(full_silhouette_space)
        if len(dfx) == 0:
            logging.info('Silhouette space empty, exiting')
            exit()

        dfx.columns = ['Silhouette', 'threshold', 'numClusters']
        sns.lineplot(x='numClusters', y='Silhouette', data=dfx, color='black')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            plt.tight_layout()
        out_path = f'{output_folder}/SilhouetteProfile.{image_format}'
        plt.savefig(out_path, dpi=300)
        plt.clf()
        plt.cla()
        logging.info('Stored the Silhouette profile.')

        final_feature_cluster_df = pd.DataFrame(list(zip(top_clustering, pivot_table.index)))
        final_feature_cluster_df.columns = ['ClusterID', 'Feature']
        final_feature_cluster_df.to_csv(f'{output_folder}/TopClustering.tsv', sep='\t')

        try:
            projected_data = TSNE().fit_transform(pivot_table.values)
            projected_data = pd.DataFrame(projected_data, columns=['Dim1', 'Dim2'])
            projected_data['ClusterID'] = top_clustering.astype(str)
            sns.scatterplot(x='Dim1', y='Dim2', hue='ClusterID', data=projected_data, palette='Set2')
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                plt.tight_layout()
            plt.savefig(f'{output_folder}/clustersEmbeddingVisualization.pdf', dpi=300)
            plt.clf()
            plt.cla()
        except:
            pass

    plt.rcParams['figure.figsize'] = (50, 30)


def visualize_heatmap(
    triplets: pd.DataFrame, output_folder: str, image_format: str,
) -> None:
    sns.set(font_scale=2)
    fig, ax = plt.subplots()
    pivot_table = pd.pivot_table(
        triplets, values='Score', index='FeatureA', columns='FeatureB', aggfunc=np.mean,
    )
    mask = np.zeros_like(pivot_table.values)
    mask[np.triu_indices_from(mask)] = True
    fsize_heatmap = 20
    if pivot_table.shape[0] > 100:
        sns.set(font_scale=1)
        fsize_heatmap = 3

    logging.info('Visualizing the heatmap ..')

    if pivot_table.shape[0] > 500:
        logging.info('Skipping heatmap visualization due to too many elements ..')
        return

    plt.figure(figsize=(50, 50))
    plt.rcParams.update({'font.size': 1})
    sns.heatmap(
        pivot_table,
        annot=True,
        mask=mask,
        annot_kws={'size': fsize_heatmap},
        square=False,
        cmap='coolwarm',
        linecolor='black',
        linewidths=0.05,
    )
    plt.xlabel('')
    plt.ylabel('')
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        plt.tight_layout()
    plt.savefig(f'{output_folder}/heatmap.{image_format}', dpi=500)
    plt.clf()
    plt.cla()
    logging.info(f'Stored heatmap to: {output_folder}/heatmap.{image_format}')


def visualize_barplots(
    triplets: pd.DataFrame,
    output_folder: str,
    reference_json: str,
    image_format: str,
    label: str,
    heuristic: str,
) -> None:
    sns.set(font_scale=8)
    feature_ranks_rows = []
    for _, row in triplets.iterrows():
        feature_A = row['FeatureA']
        feature_B = row['FeatureB']
        if label in feature_A:
            feature_ranks_rows.append([feature_B, row.Score])
        elif label in feature_B:
            feature_ranks_rows.append([feature_A, row.Score])

    feature_ranks: pd.DataFrame = pd.DataFrame(feature_ranks_rows, columns=['Feature', 'Value'])
    feature_ranks = feature_ranks[~feature_ranks['Feature'].str.contains(label)]
    if not os.path.exists(reference_json):
        reference_json = ''

    used_features = []
    if reference_json:
        ref_json = read_reference_json(reference_json)
        if 'features' in ref_json['desc']:
            used_features.extend(ref_json['desc']['features'])
        if 'fields' in ref_json['desc']:
            used_features.extend(ref_json['desc']['fields'])
    else:
        used_features = feature_ranks['Feature'].tolist()

    feature_ranks['Feature'] = feature_ranks['Feature'].astype(str)
    feature_ranks['Value'] = feature_ranks['Value'].astype(float)
    feature_ranks = feature_ranks.groupby('Feature').median().reset_index()
    feature_ranks = feature_ranks.sort_values(by='Value', ascending=False)

    subset_ranges = [10, 25, 50, 100, feature_ranks.shape[0]]
    sns.set_style('whitegrid')

    for subset_range in subset_ranges:
        feature_ranks_reduced = feature_ranks.iloc[:subset_range]
        plt.figure(figsize=(18, 12))
        fig, ax = plt.subplots()

        if 45 < feature_ranks_reduced.shape[0] <= 100:
            ax.yaxis.set_tick_params(labelsize=8)
        elif feature_ranks_reduced.shape[0] > 100:
            ax.yaxis.set_tick_params(labelsize=2)
        else:
            ax.yaxis.set_tick_params(labelsize=25)

        plt.title(f'Ranking w.r.t "{label}"\n')
        sns.barplot(
            x='Value',
            y='Feature',
            hue='Feature',
            data=feature_ranks_reduced,
            palette='coolwarm_r',
            err_kws={'linewidth': 0.7},
            dodge=False,
        )

        if ax.legend_ is not None:
            ax.legend_.remove()  # Remove the legend if it exists

        for item in ax.get_yticklabels():
            for prod_feature in used_features:
                if item.get_text() in prod_feature:
                    item.set_fontweight('bold')
                    item.set_color('red')
                    break

        plt.xlabel(f'Feature importance (based on heuristic {heuristic})')
        plt.ylabel('')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            plt.tight_layout()
        plt.savefig(f'{output_folder}/barplot_top_{subset_range}.{image_format}', dpi=300)
        plt.clf()
        plt.cla()

        logging.info(f'Stored barplot to: {output_folder}/barplot_top_{subset_range}_.{image_format}')


def visualize_all(
    triplets: pd.DataFrame,
    output_folder: str,
    label: str = '',
    reference_json: str = '',
    image_format: str = 'png',
    heuristic: str = 'MI',
) -> None:
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    visualize_hierarchical_clusters(triplets, output_folder, image_format)
    visualize_heatmap(triplets, output_folder, image_format)
    visualize_barplots(triplets, output_folder, reference_json, image_format, label, heuristic)



from __future__ import annotations

import argparse
import logging

from outrank.task_generators import outrank_task_generate_data_set
from outrank.task_instance_ranking import outrank_task_rank_instances
from outrank.task_ranking import outrank_task_conduct_ranking
from outrank.task_selftest import conduct_self_test
from outrank.task_summary import outrank_task_result_summary
from outrank.task_visualization import outrank_task_visualize_results

logging.basicConfig(
    format='%(asctime)s - %(message)s',
    datefmt='%d-%b-%y %H:%M:%S',
)
logging.getLogger(__name__).setLevel(logging.INFO)

usage_examples = """
    Usage examples:

    # perform ranking, summary and visualize the results
    outrank --task all --data_path pathToSomeData --data_source ob-vw --heuristic MI-numba-randomized --include_cardinality_in_feature_names True --target_ranking_only True --combination_number_upper_bound 2048 --num_threads 8 --interaction_order 1 --transformers fw-transformers --output_folder ./ranking_outputs --subsampling 100

    # pairwise ranking only
    outrank --task ranking --data_path pathToSomeData --data_source ob-vw --heuristic MI-numba-randomized --target_ranking_only False --combination_number_upper_bound 10000 --num_threads 30 --output_folder ./ranking_outputs --subsampling 10

    # Higher order interactions
    outrank --task all --data_path pathToSomeData --data_source csv-raw --heuristic MI-numba-randomized --target_ranking_only True --combination_number_upper_bound 2048 --num_threads 8 --interaction_order 3 --output_folder ./ranking_outputs --subsampling 20

    # More docs and use cases at https://outbrain.github.io/outrank/outrank.html
"""


def main():
    parser = argparse.ArgumentParser(
        description='Fast feature screening for sparse data sets.',
        epilog=usage_examples,
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        '--task',
        type=str,
        default='all',
        help='Type of task to consider. Can be either "ranking", "ranking_summary", "feature_summary_transformers",  or "visualization"',
    )

    parser.add_argument(
        '--minibatch_size',
        type=int,
        default=2**14,
        help='Suitable for data, not pre-split to batches, this parameter determines batch size - note that too large batch size can slow down the multithreaded score computation due to many thread allocations etc. This works ok for <300 features and up to 48 threads.',
    )

    parser.add_argument(
        '--output_folder',
        type=str,
        default='ranking_outputs',
        help='Output folder containing ranking results.',
    )

    parser.add_argument(
        '--data_source',
        type=str,
        default='ob-vw',
        help='Which database is used to obtain learning instances? this determines the inferred folder structure (csv-raw, ob-vw, ob-csv).',
    )

    parser.add_argument(
        '--data_path',
        type=str,
        default=None,
        help='Path to the folder containing the main data used for subsequent learning.',
    )

    parser.add_argument(
        '--subsampling',
        type=int,
        default=10,
        help='Subsampling ratio - every n-th instance will be considered (suggested value: 10 to 100)',
    )

    parser.add_argument(
        '--combination_number_upper_bound',
        type=int,
        default=2**15,
        help='Cap the number of columns during feature ranking, per batch. This means that if you were to evaluate e.g., 100k combinations, this parameter results in behavior where only 2 ** 15 are taken into account (randomly) each bach, resulting in a monte-carlo like sampling scheme that yields estimates of the final ranks when all data is seen.',
    )

    parser.add_argument(
        '--missing_value_symbols',
        type=str,
        default=',{}',
        help='What symbols denote missing values? Comma-separate them - if comma is a missing symbol itself please open an issue.',
    )

    parser.add_argument(
        '--heuristic',
        type=str,
        default='MI-numba-randomized',
        help='Selected heuristic (that performs feature scoring). For full list please see the docs: https://outbrain.github.io/outrank/outrank/algorithms/importance_estimator.html',
    )

    parser.add_argument(
        '--include_noise_baseline_features',
        type=str,
        default='False',
        help='If enabled, it computes five control variables (random noises)',
    )

    parser.add_argument(
        '--include_cardinality_in_feature_names',
        type=str,
        default='True',
        help='If enabled, feature names appear as feature-(cardinality) for easier inspection/debugging.',
    )

    parser.add_argument(
        '--image_format',
        type=str,
        default='pdf',
        help='The format of the output images (task: visualization)',
    )

    parser.add_argument(
        '--num_threads', type=int, default=8, help='Number of threads to consider. More threads implies faster ranking, however, there will be some memory overhead. Should be as large as the machine can handle memory-wise.',
    )

    parser.add_argument(
        '--label_column',
        type=str,
        default='label',
        help='Name of the target attribute for ranking. Note that this can be any other feature for most implemented heuristics.',
    )

    parser.add_argument(
        '--max_unique_hist_constraint',
        type=int,
        default=30_000,
        help='Max number of unique values for which counts are recalled.',
    )

    parser.add_argument(
        '--transformers',
        type=str,
        default='none',
        help='Collection of which feature transformations to consider. Examples are: fw-transformers, default',
    )

    parser.add_argument(
        '--rare_value_count_upper_bound',
        type=int,
        default=1,
        help="When identifying rare attr-val pairs, what's the upper frequency bound?",
    )

    parser.add_argument(
        '--feature_set_focus',
        type=str,
        default=None,
        help='Collection of which feature transformations to consider',
    )

    parser.add_argument(
        '--interaction_order',
        type=int,
        default=1,
        help='The order of feature interactions to consider during ranking (complex features comprised of n elementary ones)',
    )

    parser.add_argument(
        '--reference_model_JSON',
        type=str,
        default='',
        help='Reference model JSON',
    )

    parser.add_argument(
        '--target_ranking_only',
        type=str,
        default='True',
        help='Compute only the feature-label scores? This is substantially faster (O(n)).',
    )

    parser.add_argument(
        '--explode_multivalue_features',
        type=str,
        default='False',
        help="Which ';'-separated features should be one-hot encoded into n new features (coverage analysis)",
    )

    parser.add_argument(
        '--subfeature_mapping',
        type=str,
        default='False',
        help='Compute sub-features on-the fly. Example: featureA->featureB implies features based on each value of featureA will be considered. So, feature names will correspond to values of the first feature, with actual values being constructed based on the second feature (two or more possible values).',
    )

    parser.add_argument(
        '--num_synthetic_features',
        type=int,
        default=100,
        help='Relevant for task data_generator -- how many features.',
    )

    parser.add_argument(
        '--num_synthetic_rows',
        type=int,
        default=1000000,
        help='Relevant for task data_generator -- how many rows.',
    )

    parser.add_argument(
        '--generator_type',
        type=str,
        default='naive',
        help='Relevant for task data_generator -- which generator to consider',
    )

    parser.add_argument(
        '--output_synthetic_df_name',
        type=str,
        default='test_data_synthetic',
        help='Relevant for task data_generator -- name of the folder that contains generated data.',
    )

    parser.add_argument(
        '--disable_tqdm',
        default='False',
        choices=['False', 'True'],
        help='Either True or False.',
    )

    parser.add_argument(
        '--mi_stratified_sampling_ratio',
        type=float,
        default=1.0,
        help='If < 1.0, MI algorithm will further subsample data in stratified manner (equal distributions per value if possible).',
    )


    args = parser.parse_args()

    if args.task == 'selftest':
        conduct_self_test('MI-numba-randomized')
        exit()

    if args.data_path is None and args.task != 'data_generator':
        logging.error('Please specify data set name (--data_path).')
        exit()

    all_tasks_to_consider = []
    if args.task != 'all':
        all_tasks_to_consider = [args.task]

    else:
        all_tasks_to_consider = ['ranking', 'ranking_summary', 'visualization']

    for task in all_tasks_to_consider:
        logging.info(f'Proceeding with task: {task} ..')

        if (
            task == 'ranking'
            or task == 'feature_summary_transformers'
            or task == 'identify_rare_values'
        ):
            outrank_task_conduct_ranking(args)

        elif task == 'visualization':
            outrank_task_visualize_results(args)

        elif task == 'ranking_summary':
            outrank_task_result_summary(args)

        elif task == 'data_generator':
            outrank_task_generate_data_set(args)

        elif task == 'instance_ranking':
            outrank_task_rank_instances(args)

        else:
            logging.info(f'Warning, the selected task: {task} does not exist.')


if __name__ == '__main__':
    main()



from __future__ import annotations

from collections import Counter


class PrimitiveConstrainedCounter:
    """
    A memory-efficient implementation of the count min sketch algorithm with optimized hashing using Numba JIT.
    """

    def __init__(self, bound: int=(10**4) * 3):
        self.max_bound_thr = bound
        self.default_counter: Counter = Counter()

    def batch_add(self, lst):
        if len(self.default_counter) < self.max_bound_thr:
            self.default_counter = self.default_counter + Counter(lst)

    def add(self, val):
        if len(self.default_counter) < self.max_bound_thr:
            self.default_counter[val] += 1


if __name__ == '__main__':
    from collections import Counter

    depth = 8
    width = 2**22
    import numpy as np
    cms = PrimitiveConstrainedCounter()

    items = [1, 1, 2, 3, 3, 3, 4, 5, 2] * 10000
    cms.batch_add(items)  # Use the batch_add function

    print(Counter(items))  # Print the exact counts for comparison



from __future__ import annotations

import gzip
import os
from collections import Counter, defaultdict
from typing import Any

import numpy as np
import pandas as pd
import tqdm

from outrank.core_utils import generic_line_parser, get_dataset_info, get_num_of_instances

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

def shannon_entropy(string: str) -> float:
    counts = Counter(string)
    frequencies = (i / len(string) for i in counts.values())
    return -sum(f * np.log2(f) for f in frequencies)

def compute_average_entropy(line: list[str]) -> float:
    return sum(shannon_entropy(field) for field in line)

def score_line(line: list[str]) -> dict[str, float]:
    total_fields = len(line)
    nan_prop = line.count('') / total_fields
    empty_dict_prop = line.count('{}') / total_fields
    all_empty_prop = (line.count('{}') + line.count('')) / total_fields
    all_zero_prop = line.count('0') / total_fields

    out_struct = {
        'empty_string_prop': nan_prop,
        'empty_dict': empty_dict_prop,
        'all_empty': all_empty_prop,
        'all_zero': all_zero_prop,
        'row_entropy': compute_average_entropy(line)
    }

    for j in [30, 60, 100, 200, 300]:
        out_struct[f'all_more_{j}_chars'] = sum(len(x) > j for x in line) / total_fields

    return out_struct

def outrank_task_rank_instances(args: Any) -> None:
    dataset_info = get_dataset_info(args)
    data_path = dataset_info.data_path
    data_encoding = 'utf-8'
    delimiter = '\t'

    total_lines = get_num_of_instances(data_path) - 1
    local_pbar = tqdm.tqdm(total=total_lines, position=0, disable=args.disable_tqdm == 'True')
    local_pbar.set_description('Starting ranking computation')

    _, file_extension = os.path.splitext(data_path)
    file_stream = gzip.open(data_path, 'rt', encoding=data_encoding) if file_extension == '.gz' else open(data_path, encoding=data_encoding)

    line_counter = 0
    out_scores_lab = defaultdict(list)

    for line in file_stream:
        line_counter += 1
        local_pbar.update(1)

        parsed_line = generic_line_parser(
            line,
            delimiter,
            args,
            dataset_info.fw_map,
            dataset_info.column_names,
        )

        if line_counter > 100_000:
            break
        out_scores_lab[line[0]].append(score_line(parsed_line))

    file_stream.close()

    os.makedirs(args.output_folder, exist_ok=True)
    for label, out_scores in out_scores_lab.items():
        out_df = pd.DataFrame(out_scores)
        for col in out_df.columns:
            sorted_vals = out_df[col].sort_values()
            plt.figure(figsize=(5, 5), dpi=300)
            plt.title(f'{col} label: {label}')
            plt.hist(
                x=sorted_vals * 100,
                color='black',
                density=True,
                bins=100,
            )
            plt.xlabel('Proportion of namespaces (%)' if 'entropy' not in col else 'Row entropy')
            plt.ylabel('Density')
            plt.tight_layout()
            fname = f'distPlot{col}_{label}.pdf'
            plt.savefig(os.path.join(args.output_folder, fname), dpi=300)
            plt.cla()
            plt.clf()

    local_pbar.close()



from __future__ import annotations

import sys
from collections import Counter

import numpy as np
from numba import njit
from numba import prange


@njit
def cms_hash(x, seed, width):
    x_hash = np.uint32(hash(x))
    return (x_hash + seed) % width

class CountMinSketch:
    """
    A memory-efficient implementation of the count min sketch algorithm with optimized hashing using Numba JIT.
    """

    def __init__(self, depth=6, width=2**15, M=None):
        self.depth = depth
        self.width = width
        self.hash_seeds = np.array(np.random.randint(low=0, high=2**31 - 1, size=depth), dtype=np.uint32)
        self.M = np.zeros((depth, width), dtype=np.int32) if M is None else M

    @staticmethod
    @njit
    def _add(M, x, depth, width, hash_seeds, delta=1):
        for i in prange(depth):
            location = cms_hash(x, hash_seeds[i], width)
            M[i, location] += delta

    def add(self, x, delta=1):
        CountMinSketch._add(self.M, x, self.depth, self.width, self.hash_seeds, delta)

    def batch_add(self, lst, delta=1):
        for x in lst:
            self.add(x, delta)

    def query(self, x):
        return min(self.M[i][cms_hash(x, self.hash_seeds[i], self.width)] for i in range(self.depth))

    def get_matrix(self):
        return self.M


if __name__ == '__main__':
    from collections import Counter

    depth = 8
    width = 2**22
    cms = CountMinSketch(depth, width)

    items = [1, 1, 2, 3, 3, 3, 4, 5, 2] * 1000
    cms.batch_add(items)  # Use the batch_add function

    print(cms.query(3))  # Query for frequency estimates
    print(cms.query(1))
    print(cms.query(2))
    print(cms.query(4))
    print(cms.query(5))

    print(Counter(items))  # Print the exact counts for comparison



# Some boilerplate transformations people tend to use
from __future__ import annotations

MINIMAL_TRANSFORMERS = {
    '_tr_sqrt': 'np.sqrt(X)',
    '_tr_log(x+1)': 'np.log(X + 1)',
    '_tr_sqrt(abs(x))': 'np.sqrt(np.abs(X))',
    '_tr_log(abs(x)+1)': 'np.log(np.abs(X) + 1)',
}

DEFAULT_TRANSFORMERS = {
    '_tr_sqrt': 'np.sqrt(X)',
    '_tr_log(x+1)': 'np.log(X + 1)',
    '_tr_sqrt(abs(x))': 'np.sqrt(np.abs(X))',
    '_tr_log(abs(x)+1)': 'np.log(np.abs(X) + 1)',
    '_tr_div(x,abs(x))*log(abs(x))': 'np.divide(X, np.abs(X)) * np.log(np.abs(X))',
    '_tr_log(x + sqrt(pow(x,2), 1)': 'np.log(X + np.sqrt(np.power(X, 2) + 1))',
    '_tr_log*sqrt': 'np.log(X + 1) * np.sqrt(X)',
    '_tr_log*100': 'np.round(np.log(X + 1) * 100, 0)',
    '_tr_nonzero': 'np.where(X != 0, 1, 0)',
    '_tr_round(div(x,max))': 'np.round(np.divide(X, np.max(X)), 0)',
}

EXTENDED_TRANSFORMERS = {
    '_tr_sqrt': 'np.sqrt(X)',
    '_tr_log(x+1)': 'np.log(X + 1)',
    '_tr_sqrt(abs(x))': 'np.sqrt(np.abs(X))',
    '_tr_log(abs(x)+1)': 'np.log(np.abs(X) + 1)',
    '_tr_div(x,abs(x))*log(abs(x))': 'np.divide(X, np.abs(X)) * np.log(np.abs(X))',
    '_tr_log(x + sqrt(pow(x,2), 1)': 'np.log(X + np.sqrt(np.power(X, 2) + 1))',
    '_tr_log*sqrt': 'np.log(X + 1) * np.sqrt(X)',
    '_tr_log*100': 'np.round(np.log(X + 1) * 100, 0)',
    '_tr_nonzero': 'np.where(X != 0, 1, 0)',
    '_tr_round(div(x,max))': 'np.round(np.divide(X, np.max(X)), 0)',
    '_tr_squared': 'np.square(X)',
    '_tr_cubed': 'np.power(X, 3)',
    '_tr_inverse': 'np.divide(1, X, out=np.zeros_like(X), where=X!=0)',
    '_tr_exp': 'np.exp(X)',
    '_tr_exp_minus_1': 'np.expm1(X)',
    '_tr_sin': 'np.sin(X)',
    '_tr_cos': 'np.cos(X)',
    '_tr_tan': 'np.tan(X)',
    '_tr_arcsin_safe': 'np.arcsin(np.clip(X, -1, 1))',
    '_tr_arccos_safe': 'np.arccos(np.clip(X, -1, 1))',
    '_tr_arctan': 'np.arctan(X)',
    '_tr_sign': 'np.sign(X)',
    '_tr_z_score': '(X - np.mean(X)) / np.std(X)',
    '_tr_min_max_scale': '(X - np.min(X)) / (np.max(X) - np.min(X))',
    '_tr_quadratic_root': 'np.cbrt(X)',
    '_tr_shifted_sigmoid': '1 / (1 + np.exp(-X - 0.5))',
    '_tr_logistic': '1 / (1 + np.exp(-X))',
}

VERBOSE_TRANSFORMERS = {
    '_tr_sqrt': 'np.sqrt(X)',
    '_tr_log(x+1)': 'np.log(X + 1)',
    '_tr_sqrt(abs(x))': 'np.sqrt(np.abs(X))',
    '_tr_log(abs(x)+1)': 'np.log(np.abs(X) + 1)',
    '_tr_div(x,abs(x))*log(abs(x))': 'np.divide(X, np.abs(X)) * np.log(np.abs(X))',
    '_tr_log(x + sqrt(pow(x,2), 1)': 'np.log(X + np.sqrt(np.power(X, 2) + 1))',
    '_tr_log*sqrt': 'np.log(X + 1) * np.sqrt(X)',
    '_tr_log*100': 'np.round(np.log(X + 1) * 100, 0)',
    '_tr_nonzero': 'np.where(X != 0, 1, 0)',
    '_tr_round(div(x,max))': 'np.round(np.divide(X, np.max(X)), 0)',
    '_tr_squared': 'np.square(X)',
    '_tr_cubed': 'np.power(X, 3)',
    '_tr_inverse': 'np.divide(1, X, out=np.zeros_like(X), where=X!=0)',
    '_tr_exp': 'np.exp(X)',
    '_tr_exp_minus_1': 'np.expm1(X)',
    '_tr_sin': 'np.sin(X)',
    '_tr_cos': 'np.cos(X)',
    '_tr_tan': 'np.tan(X)',
    '_tr_arcsin_safe': 'np.arcsin(np.clip(X, -1, 1))',
    '_tr_arccos_safe': 'np.arccos(np.clip(X, -1, 1))',
    '_tr_arctan': 'np.arctan(X)',
    '_tr_sign': 'np.sign(X)',
    '_tr_z_score': '(X - np.mean(X)) / np.std(X)',
    '_tr_min_max_scale': '(X - np.min(X)) / (np.max(X) - np.min(X))',
    '_tr_quadratic_root': 'np.cbrt(X)',
    '_tr_shifted_sigmoid': '1 / (1 + np.exp(-X - 0.5))',
    '_tr_logistic': '1 / (1 + np.exp(-X))',
    '_tr_reciprocal': 'np.reciprocal(X, out=np.zeros_like(X), where=X!=0)',
    '_tr_log1p': 'np.log1p(X)',
    '_tr_expm1': 'np.expm1(X)',
    '_tr_softplus': 'np.log1p(np.exp(X))',
    '_tr_absolute': 'np.abs(X)',
    '_tr_clip_0_1': 'np.clip(X, 0, 1)',
    '_tr_normalize': '(X - np.min(X)) / (np.max(X) - np.min(X))',
    '_tr_standardize': '(X - np.mean(X)) / np.std(X)',
    '_tr_cumulative_sum': 'np.cumsum(X)',
    '_tr_cumulative_prod': 'np.cumprod(X)',
    '_tr_diff': 'np.diff(X, prepend=0)',
    '_tr_fourth_root': 'np.sqrt(np.sqrt(X))',
    '_tr_log10(x+1)': 'np.log10(X + 1)',
    '_tr_fourth_root(abs(x))': 'np.sqrt(np.sqrt(np.abs(X)))',
    '_tr_log10(abs(x)+1)': 'np.log10(np.abs(X) + 1)',
    '_tr_div(x,abs(x))*log10(abs(x))': 'np.divide(X, np.abs(X)) * np.log10(np.abs(X))',
    '_tr_log10*sqrt': 'np.log10(X + 1) * np.sqrt(X)',
    '_tr_log10*100': 'np.round(np.log10(X + 1) * 100, 0)',
    '_tr_positive': 'np.where(X > 0, 1, 0)',
    '_tr_round(div(x,median))': 'np.round(np.divide(X, np.median(X)), 0)',
    '_tr_fourth_power': 'np.power(X, 4)',
    '_tr_fifth_power': 'np.power(X, 5)',
    '_tr_inverse_squared': 'np.divide(1, np.square(X), out=np.zeros_like(X), where=X!=0)',
    '_tr_exp2': 'np.exp2(X)',
    '_tr_sinh': 'np.sinh(X)',
    '_tr_cosh': 'np.cosh(X)',
    '_tr_tanh': 'np.tanh(X)',
    '_tr_log2': 'np.log2(X)',
    '_tr_log10': 'np.log10(X)',
    '_tr_sinc': 'np.sinc(X)',
    '_tr_mean': 'np.mean(X)',
    '_tr_median': 'np.median(X)',
    '_tr_variance': 'np.var(X)',
    '_tr_std_dev': 'np.std(X)',
    '_tr_percentile_25': 'np.percentile(X, 25)',
    '_tr_percentile_50': 'np.percentile(X, 50)',
    '_tr_percentile_75': 'np.percentile(X, 75)',
    '_tr_reverse': 'X[::-1]',
    '_tr_sort': 'np.sort(X)',
    '_tr_argsort': 'np.argsort(X)',
    '_tr_is_nan': 'np.isnan(X)',
    '_tr_is_finite': 'np.isfinite(X)',
    '_tr_is_infinite': 'np.isinf(X)',
    '_tr_is_positive': 'np.greater(X, 0)',
    '_tr_is_negative': 'np.less(X, 0)',
    '_tr_is_zero': 'np.equal(X, 0)',
    '_tr_log2(x+1)': 'np.log2(X + 1)',
    '_tr_log2(abs(x)+1)': 'np.log2(np.abs(X) + 1)',
    '_tr_log2*sqrt': 'np.log2(X + 1) * np.sqrt(X)',
    '_tr_log2*100': 'np.round(np.log2(X + 1) * 100, 0)',
    '_tr_sec': '1 / np.cos(X)',
    '_tr_cosec': '1 / np.sin(X)',
    '_tr_cot': '1 / np.tan(X)',
    '_tr_arcsinh': 'np.arcsinh(X)',
    '_tr_arccosh': 'np.arccosh(X)',
    '_tr_arctanh': 'np.arctanh(X)',
    '_tr_poly2': 'np.polyval([1, 0, 0], X)',
    '_tr_poly3': 'np.polyval([1, 0, 0, 0], X)',
    '_tr_poly4': 'np.polyval([1, 0, 0, 0, 0], X)',
    '_tr_custom1': 'X ** 2 + 2 * X + 1',
    '_tr_custom2': 'np.sin(X) + np.cos(X)',
    '_tr_custom3': 'np.exp(X) - np.log(X + 1)',
    '_tr_empirical_cdf': 'np.searchsorted(np.sort(X), X, side="right") / len(X)',
    '_tr_uniform_cdf': 'np.clip(X, 0, 1)',
    '_tr_exponential_cdf': '1 - np.exp(-X)',
    '_tr_square_root_abs': 'np.sqrt(np.abs(X))',
    '_tr_log1p_abs': 'np.log1p(np.abs(X))',
    '_tr_exp_minus_one': 'np.expm1(X)',
    '_tr_inverse_abs': 'np.divide(1, np.abs(X), out=np.zeros_like(X), where=X!=0)',
    '_tr_square_root_inverse': 'np.sqrt(np.divide(1, X, out=np.zeros_like(X), where=X!=0))',
    '_tr_cubed_root': 'np.cbrt(X)',
    '_tr_log1p_squared': 'np.log1p(np.square(X))',
    '_tr_exp_squared': 'np.exp(np.square(X))',
    '_tr_sin_squared': 'np.sin(np.square(X))',
    '_tr_cos_squared': 'np.cos(np.square(X))',
    '_tr_tan_squared': 'np.tan(np.square(X))',
    '_tr_arctan_squared': 'np.arctan(np.square(X))',
    '_tr_sign_squared': 'np.sign(np.square(X))',
    '_tr_z_score_abs': '(np.abs(X) - np.mean(np.abs(X))) / np.std(np.abs(X))',
    '_tr_min_max_scale_abs': '(np.abs(X) - np.min(np.abs(X))) / (np.max(np.abs(X)) - np.min(np.abs(X)))',
    '_tr_shifted_sigmoid_abs': '1 / (1 + np.exp(-np.abs(X) - 0.5))',
    '_tr_logistic_abs': '1 / (1 + np.exp(-np.abs(X)))',
    '_tr_reciprocal_abs': 'np.reciprocal(np.abs(X), out=np.zeros_like(X), where=X!=0)',
    '_tr_log1p_cubed': 'np.log1p(np.power(X, 3))',
    '_tr_expm1_cubed': 'np.expm1(np.power(X, 3))',
    '_tr_softplus_cubed': 'np.log1p(np.exp(np.power(X, 3)))',
    '_tr_absolute_squared': 'np.abs(np.square(X))',
    '_tr_clip_0_2': 'np.clip(X, 0, 2)',
    '_tr_normalize_abs': '(np.abs(X) - np.min(np.abs(X))) / (np.max(np.abs(X)) - np.min(np.abs(X)))',
    '_tr_standardize_abs': '(np.abs(X) - np.mean(np.abs(X))) / np.std(np.abs(X))',
    '_tr_cumulative_sum_abs': 'np.cumsum(np.abs(X))',
    '_tr_cumulative_prod_abs': 'np.cumprod(np.abs(X))',
    '_tr_diff_abs': 'np.diff(np.abs(X), prepend=0)',
    '_tr_fourth_root_abs': 'np.sqrt(np.sqrt(np.abs(X)))',
    '_tr_log10_abs(x+1)': 'np.log10(np.abs(X) + 1)',
    '_tr_fourth_root_squared': 'np.sqrt(np.sqrt(np.square(X)))',
    '_tr_log10_squared(abs(x)+1)': 'np.log10(np.square(np.abs(X)) + 1)',
    '_tr_div_squared(x,abs(x))*log10(abs(x))': 'np.divide(np.square(X), np.abs(X)) * np.log10(np.abs(X))',
    '_tr_log10*sqrt_squared': 'np.log10(np.square(X) + 1) * np.sqrt(np.square(X))',
    '_tr_log10*100_squared': 'np.round(np.log10(np.square(X) + 1) * 100, 0)',
    '_tr_positive_squared': 'np.where(np.square(X) > 0, 1, 0)',
    '_tr_round_squared(div(x,median))': 'np.round(np.divide(np.square(X), np.median(np.square(X))), 0)',
    '_tr_fourth_power_squared': 'np.power(np.square(X), 4)',
    '_tr_fifth_power_squared': 'np.power(np.square(X), 5)',
    '_tr_inverse_squared_squared': 'np.divide(1, np.square(np.square(X)), out=np.zeros_like(X), where=X!=0)',
    '_tr_exp2_squared': 'np.exp2(np.square(X))',
    '_tr_sinh_squared': 'np.sinh(np.square(X))',
    '_tr_cosh_squared': 'np.cosh(np.square(X))',
    '_tr_tanh_squared': 'np.tanh(np.square(X))',
    '_tr_log2_squared': 'np.log2(np.square(X))',
    '_tr_log10_squared': 'np.log10(np.square(X))',
    '_tr_sinc_squared': 'np.sinc(np.square(X))',
    '_tr_mean_squared': 'np.mean(np.square(X))',
    '_tr_median_squared': 'np.median(np.square(X))',
    '_tr_variance_squared': 'np.var(np.square(X))',
    '_tr_std_dev_squared': 'np.std(np.square(X))',
    '_tr_percentile_25_squared': 'np.percentile(np.square(X), 25)',
    '_tr_percentile_50_squared': 'np.percentile(np.square(X), 50)',
    '_tr_percentile_75_squared': 'np.percentile(np.square(X), 75)',
    '_tr_reverse_squared': 'np.square(X)[::-1]',
    '_tr_sort_squared': 'np.sort(np.square(X))',
    '_tr_argsort_squared': 'np.argsort(np.square(X))',
    '_tr_is_nan_squared': 'np.isnan(np.square(X))',
    '_tr_is_finite_squared': 'np.isfinite(np.square(X))',
    '_tr_is_infinite_squared': 'np.isinf(np.square(X))',
    '_tr_is_positive_squared': 'np.greater(np.square(X), 0)',
    '_tr_is_negative_squared': 'np.less(np.square(X), 0)',
    '_tr_is_zero_squared': 'np.equal(np.square(X), 0)',
    '_tr_log2_squared(x+1)': 'np.log2(np.square(X) + 1)',
    '_tr_log2_squared(abs(x)+1)': 'np.log2(np.square(np.abs(X)) + 1)',
    '_tr_log2*sqrt_squared': 'np.log2(np.square(X) + 1) * np.sqrt(np.square(X))',
    '_tr_log2*100_squared': 'np.round(np.log2(np.square(X) + 1) * 100, 0)',
    '_tr_sec_squared': '1 / np.cos(np.square(X))',
    '_tr_cosec_squared': '1 / np.sin(np.square(X))',
    '_tr_cot_squared': '1 / np.tan(np.square(X))',
    '_tr_arcsinh_squared': 'np.arcsinh(np.square(X))',
    '_tr_arccosh_squared': 'np.arccosh(np.square(X))',
    '_tr_arctanh_squared': 'np.arctanh(np.square(X))',
    '_tr_poly2_squared': 'np.polyval([1, 0, 0], np.square(X))',
    '_tr_poly3_squared': 'np.polyval([1, 0, 0, 0], np.square(X))',
    '_tr_poly4_squared': 'np.polyval([1, 0, 0, 0, 0], np.square(X))',
    '_tr_custom1_squared': 'np.square(X) ** 2 + 2 * np.square(X) + 1',
    '_tr_custom2_squared': 'np.sin(np.square(X)) + np.cos(np.square(X))',
    '_tr_custom3_squared': 'np.exp(np.square(X)) - np.log(np.square(X) + 1)',
    '_tr_empirical_cdf_squared': 'np.searchsorted(np.sort(np.square(X)), np.square(X), side="right") / len(np.square(X))',
    '_tr_uniform_cdf_squared': 'np.clip(np.square(X), 0, 1)',
    '_tr_exponential_cdf_squared': '1 - np.exp(-np.square(X))',
    '_tr_fill_nan_with_zero': 'np.nan_to_num(X, nan=0.0)',
    '_tr_fill_nan_with_mean': 'np.where(np.isnan(X), np.nanmean(X), X)',
    '_tr_fill_nan_with_median': 'np.where(np.isnan(X), np.nanmedian(X), X)',
    '_tr_fill_nan_with_max': 'np.where(np.isnan(X), np.nanmax(X), X)',
    '_tr_fill_nan_with_min': 'np.where(np.isnan(X), np.nanmin(X), X)',
    '_tr_fill_inf_with_zero': 'np.where(np.isinf(X), 0.0, X)',
    '_tr_fill_inf_with_mean': 'np.where(np.isinf(X), np.nanmean(X), X)',
    '_tr_fill_inf_with_median': 'np.where(np.isinf(X), np.nanmedian(X), X)',
    '_tr_fill_inf_with_max': 'np.where(np.isinf(X), np.nanmax(X), X)',
    '_tr_fill_inf_with_min': 'np.where(np.isinf(X), np.nanmin(X), X)',
    '_tr_fill_zero_with_nan': 'np.where(X == 0, np.nan, X)',
    '_tr_fill_zero_with_mean': 'np.where(X == 0, np.nanmean(X), X)',
    '_tr_fill_zero_with_median': 'np.where(X == 0, np.nanmedian(X), X)',
    '_tr_fill_zero_with_max': 'np.where(X == 0, np.nanmax(X), X)',
    '_tr_fill_zero_with_min': 'np.where(X == 0, np.nanmin(X), X)',
    '_tr_fill_negative_with_zero': 'np.where(X < 0, 0, X)',
    '_tr_fill_negative_with_nan': 'np.where(X < 0, np.nan, X)',
    '_tr_fill_negative_with_mean': 'np.where(X < 0, np.nanmean(X), X)',
    '_tr_fill_negative_with_median': 'np.where(X < 0, np.nanmedian(X), X)',
    '_tr_fill_negative_with_max': 'np.where(X < 0, np.nanmax(X), X)',
    '_tr_fill_negative_with_min': 'np.where(X < 0, np.nanmin(X), X)',
    '_tr_fill_positive_with_zero': 'np.where(X > 0, 0, X)',
    '_tr_fill_positive_with_nan': 'np.where(X > 0, np.nan, X)',
    '_tr_fill_positive_with_mean': 'np.where(X > 0, np.nanmean(X), X)',
    '_tr_fill_positive_with_median': 'np.where(X > 0, np.nanmedian(X), X)',
    '_tr_fill_positive_with_max': 'np.where(X > 0, np.nanmax(X), X)',
    '_tr_fill_positive_with_min': 'np.where(X > 0, np.nanmin(X), X)',
    '_tr_threshold_10_with_max': 'np.where(X > 10, np.nanmax(X), X)',
    '_tr_threshold_10_with_min': 'np.where(X > 10, np.nanmin(X), X)',
    '_tr_threshold_10_with_mean': 'np.where(X > 10, np.nanmean(X), X)',
    '_tr_threshold_10_with_median': 'np.where(X > 10, np.nanmedian(X), X)',
    '_tr_threshold_minus_10_with_zero': 'np.where(X < -10, 0, X)',
    '_tr_threshold_minus_10_with_nan': 'np.where(X < -10, np.nan, X)',
    '_tr_threshold_minus_10_with_mean': 'np.where(X < -10, np.nanmean(X), X)',
    '_tr_threshold_minus_10_with_median': 'np.where(X < -10, np.nanmedian(X), X)',
    '_tr_clip_outliers_1_99': 'np.clip(X, np.percentile(X, 1), np.percentile(X, 99))',
    '_tr_clip_outliers_5_95': 'np.clip(X, np.percentile(X, 5), np.percentile(X, 95))',
    '_tr_clip_outliers_10_90': 'np.clip(X, np.percentile(X, 10), np.percentile(X, 90))',
    '_tr_clip_outliers_25_75': 'np.clip(X, np.percentile(X, 25), np.percentile(X, 75))',
    '_tr_boxcox_lambda_0.5': 'np.where(X > 0, (X ** 0.5 - 1) / 0.5, np.nan)',
    '_tr_yeo_johnson_lambda_0': 'np.sign(X) * np.log(np.abs(X) + 1)',
    '_tr_relu': 'np.maximum(0, X)',
    '_tr_leaky_relu': 'np.where(X > 0, X, 0.01 * X)',
    '_tr_elu': 'np.where(X > 0, X, np.exp(X) - 1)',
    '_tr_softsign': 'X / (1 + np.abs(X))',
    '_tr_swish': 'X / (1 + np.exp(-X))',
    '_tr_gelu_approx': '0.5 * X * (1 + np.tanh(np.sqrt(2 / np.pi) * (X + 0.044715 * X ** 3)))',
    '_tr_mish': 'X * np.tanh(np.log1p(np.exp(X)))',
    '_tr_hard_sigmoid': 'np.clip((X + 1) / 2, 0, 1)',
    '_tr_hard_tanh': 'np.clip(X, -1, 1)',
    '_tr_swish_flick': 'X / (1 + np.exp(X))',
    '_tr_reciprocal_softplus': '1 / np.log1p(np.exp(X))',
    '_tr_bell_shape': 'np.exp(-X ** 2)',
    '_tr_gumbel_smx': 'np.exp(-np.exp(-(X - 3.0) / 4.0))',
    '_tr_gumbel_smx1': 'np.exp(-np.exp(-(X - 0.5) / 2.0))',
    '_tr_gumbel_smx2': 'np.exp(-np.exp(-(X - 2.0) / 1.2))',
    '_tr_fancy_beta': 'np.sqrt(2 * np.pi / X) * (X / np.e) ** X',
    '_tr_energy_distance': 'np.sqrt(2 * np.var(X))',
}


EXTENDED_ROUNDED_TRANSFORMERS = {}
for k, v in EXTENDED_TRANSFORMERS.items():
    for round_param in [1, 2, 3, 4]:
        new_key = k + f'_round{round_param}'
        new_value = f'np.round(np.astype(np.array(' + v + f'), np.float32), {round_param})'
        EXTENDED_ROUNDED_TRANSFORMERS[new_key] = new_value


if __name__ == '__main__':
    import numpy as np

    # generate some input (call it X)
    X = np.random.random(100)

    # get some transformer
    some_transformer = DEFAULT_TRANSFORMERS.get('_tr_nonzero')

    if some_transformer is None:
        some_transformer = ''

    # evaluate to get output
    output = eval(some_transformer)

    # check output somehow
    print(output)



from __future__ import annotations

import logging
import operator
import traceback
from typing import Any, Dict, List, Tuple

import numpy as np
import pandas as pd
from scipy.stats import pearsonr
from sklearn.feature_selection import mutual_info_classif
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.metrics import adjusted_mutual_info_score
from sklearn.model_selection import cross_val_score
from sklearn.preprocessing import OneHotEncoder
from sklearn.svm import SVC

from outrank.algorithms.feature_ranking import ranking_cov_alignment
from outrank.core_utils import is_prior_heuristic

logger = logging.getLogger('syn-logger')
logger.setLevel(logging.DEBUG)

num_folds = 4

try:
    from outrank.algorithms.feature_ranking import ranking_mi_numba
    numba_available = True
except ImportError:
    traceback.print_exc()
    numba_available = False

def sklearn_MI(vector_first: np.ndarray, vector_second: np.ndarray) -> float:
    return mutual_info_classif(
        vector_first.reshape(-1, 1), vector_second.reshape(-1), discrete_features=True
    )[0]

def sklearn_surrogate(
    vector_first: np.ndarray, vector_second: np.ndarray, X: np.ndarray, surrogate_model: str
) -> float:
    clf = initialize_classifier(surrogate_model)
    transf = OneHotEncoder()

    if len(np.unique(vector_second)) > 2:
        vector_first, vector_second = vector_second, vector_first

    if X.size <= 1:
        X = vector_first.reshape(-1, 1)
    else:
        X = np.concatenate((X, vector_first.reshape(-1, 1)), axis=1)

    X = transf.fit_transform(X)
    scores = cross_val_score(clf, X, vector_second, scoring='neg_log_loss', cv=num_folds)
    return 1 + np.median(scores)

def numba_mi(vector_first: np.ndarray, vector_second: np.ndarray, heuristic: str, mi_stratified_sampling_ratio: float) -> float:
    cardinality_correction = heuristic == 'MI-numba-randomized'
    return ranking_mi_numba.mutual_info_estimator_numba(
        vector_first.astype(np.int32),
        vector_second.astype(np.int32),
        approximation_factor=np.float32(mi_stratified_sampling_ratio),
        cardinality_correction=cardinality_correction,
    )

def sklearn_mi_adj(vector_first: np.ndarray, vector_second: np.ndarray) -> float:
    return adjusted_mutual_info_score(vector_first, vector_second)

def get_importances_estimate_pairwise(combination: Tuple[str, str], reference_model_features: List[str], args: Any, tmp_df: pd.DataFrame) -> Tuple[str, str, float]:
    feature_one, feature_two = combination

    if feature_one not in tmp_df.columns or feature_two not in tmp_df.columns:
        logger.info(f'{feature_one} or {feature_two} not found in the constructed data frame.')
        return feature_one, feature_two, 0.0

    vector_first = tmp_df[feature_one].values
    vector_second = tmp_df[feature_two].values

    if vector_first.size == 0 or vector_second.size == 0:
        return feature_one, feature_two, 0.0

    if args.heuristic == 'MI':
        score = sklearn_MI(vector_first, vector_second)
    elif 'surrogate-' in args.heuristic:
        X = tmp_df[reference_model_features].values if is_prior_heuristic(args) and reference_model_features else np.array([])
        score = sklearn_surrogate(vector_first, vector_second, X, args.heuristic)
    elif 'max-value-coverage' in args.heuristic:
        score = ranking_cov_alignment.max_pair_coverage(vector_first, vector_second)
    elif 'MI-numba' in args.heuristic:
        score = numba_mi(vector_first, vector_second, args.heuristic, args.mi_stratified_sampling_ratio)
    elif args.heuristic == 'AMI':
        score = sklearn_mi_adj(vector_first, vector_second)
    elif args.heuristic == 'correlation-Pearson':
        score = pearsonr(vector_first, vector_second)[0]
    elif args.heuristic == 'Constant':
        score = 0.0
    else:
        raise ValueError('Please select a valid heuristic (MI, chi2, etc.).')

    return feature_one, feature_two, score

def rank_features_3MR(
    relevance_dict: Dict[str, float],
    redundancy_dict: Dict[Tuple[Any, Any], Any],
    relational_dict: Dict[Tuple[Any, Any], Any],
    strategy: str = 'median',
    alpha: float = 1.0,
    beta: float = 1.0,
) -> pd.DataFrame:
    all_features = set(relevance_dict.keys())
    most_important_feature = max(relevance_dict.items(), key=operator.itemgetter(1))[0]
    ranked_features = [most_important_feature]

    def calc_higher_order(feature: str, is_redundancy: bool = True) -> float:
        values = []
        for feat in ranked_features:
            interaction_tuple = (feat, feature)
            if is_redundancy:
                values.append(redundancy_dict.get(interaction_tuple, 0))
            else:
                values.append(relational_dict.get(interaction_tuple, 0))
        return np.median(values) if strategy == 'median' else (np.mean(values) if strategy == 'mean' else sum(values))

    while len(ranked_features) < len(all_features):
        top_importance = -np.inf
        most_important_feature = None

        for feat in all_features - set(ranked_features):
            feature_redundancy = calc_higher_order(feat)
            feature_relation = calc_higher_order(feat, False)
            feature_relevance = relevance_dict[feat]
            importance = feature_relevance - alpha * feature_redundancy + beta * feature_relation

            if importance > top_importance:
                top_importance = importance
                most_important_feature = feat

        ranked_features.append(most_important_feature)

    return pd.DataFrame({'Feature': ranked_features, '3MR_Ranking': range(1, len(ranked_features) + 1)})

def get_importances_estimate_nonmyopic(args: Any, tmp_df: pd.DataFrame):
    pass

def initialize_classifier(surrogate_model: str):
    if 'surrogate-LR' in surrogate_model:
        return LogisticRegression(max_iter=100000)
    elif 'surrogate-SVM' in surrogate_model:
        return SVC(gamma='auto', probability=True)
    elif 'surrogate-SGD' in surrogate_model:
        return SGDClassifier(max_iter=100000, loss='log_loss')
    else:
        logger.warning(f'The chosen surrogate model {surrogate_model} is not supported, falling back to surrogate-SGD')
        return SGDClassifier(max_iter=100000, loss='log_loss')



from __future__ import annotations

from typing import List
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from numpy.typing import ArrayLike
from scipy.linalg import qr
from scipy.stats import norm
from sklearn.cluster import KMeans
from sklearn.utils import resample


class CategoricalClassification:

    def __init__(self, seed: int = 42):
        np.random.seed(seed)
        self.dataset_info = {
            'general': {},
            'combinations': [],
            'correlations': [],
            'duplicates': [],
            'labels': {},
            'noise': [],
        }

    def __repr__(self):
        return f"CategoricalClassification(dataset_info={self.dataset_info})"

    def generate_data(
        self,
        n_features: int,
        n_samples: int,
        cardinality: int = 5,
        structure: list | ArrayLike | None = None,
        ensure_rep: bool = False,
        random_values: bool | None = False,
        low: int | None = 0,
        high: int | None = 1000,
        k: int | float = 10,
        seed: int = 42,
    ) -> np.ndarray:

        """
        Generates dataset based on given parameters
        :param n_features: number of generated features
        :param n_samples: number of generated samples
        :param cardinality: default cardinality of the dataset
        :param structure: structure of the dataset
        :param ensure_rep: flag, ensures all given values represented
        :param random_values: flag, enables random (integer) feature values from set [low, high]
        :param low: sets lower bound of random feature values
        :param high: sets high bound of random feature values
        :param k: scale constant for normal distribution, default 10, sets width of normal distribution, larger value -> narrower peak
        :param seed: sets seed of numpy random
        :return: X, 2D dataset
        """

        self.dataset_info.update({
            'general': {
                'n_features': n_features,
                'n_samples': n_samples,
                'cardinality': cardinality,
                'structure': structure,
                'ensure_rep': ensure_rep,
                'seed': seed,
            },
        })

        np.random.seed(seed)
        X = np.empty([n_features, n_samples], dtype='int32')

        # No specific structure parameter passed
        if structure is None:
            for i in range(n_features):
                x = self._generate_feature(
                    n_samples,
                    cardinality=cardinality,
                    ensure_rep=ensure_rep,
                    random_values=random_values,
                    low=low,
                    high=high,
                    k=k,
                )
                X[i] = x
        # Structure parameter passed, building based on structure
        else:
            ix = 0
            for data in structure:

                # Data in structure is a tuple of (feature index (integer), feature attributes)
                if not isinstance(data[0], (list, np.ndarray)):
                    feature_ix, feature_attributes = data

                    # Filling out the dataset up to column index feature_ix
                    if ix < feature_ix:
                        for i in range(ix, feature_ix):
                            x = self._generate_feature(
                                n_samples,
                                cardinality=cardinality,
                                ensure_rep=ensure_rep,
                                random_values=random_values,
                                low=low,
                                high=high,
                                k=k,
                            )
                            X[ix] = x
                            ix += 1

                    x = self._configure_generate_feature(
                        feature_attributes,
                        n_samples,
                        ensure_rep=ensure_rep,
                        random_values=random_values,
                        low=low,
                        high=high,
                    )
                    X[ix] = x
                    ix += 1

                # Data in structure is a tuple of (list of feature indexes, feature attributes)
                else:
                    feature_ixs, feature_attributes = data

                    # Filling out the dataset up to feature_ix
                    for feature_ix in feature_ixs:
                        if ix < feature_ix:
                            for i in range(ix, feature_ix):
                                x = self._generate_feature(
                                    n_samples,
                                    cardinality=cardinality,
                                    ensure_rep=ensure_rep,
                                    random_values=random_values,
                                    low=low,
                                    high=high,
                                    k=k,
                                )
                                X[ix] = x
                                ix += 1

                        x = self._configure_generate_feature(
                            feature_attributes,
                            n_samples,
                            ensure_rep=ensure_rep,
                            random_values=random_values,
                            low=low,
                            high=high,
                        )

                        X[ix] = x
                        ix += 1

            # Fill out the rest of the dataset
            if ix < n_features:
                for i in range(ix, n_features):
                    x = self._generate_feature(
                        n_samples,
                        cardinality=cardinality,
                        ensure_rep=ensure_rep,
                        random_values=random_values,
                        low=low,
                        high=high,
                        k=k,
                    )
                    X[i] = x

        return X.T

    def _configure_generate_feature(
        self,
        feature_attributes: int | list | ArrayLike,
        n_samples: int,
        ensure_rep: bool = False,
        random_values: bool | None = False,
        low: int | None = 0,
        high: int | None = 1000,
        k: int | float = 10,
    ) -> np.ndarray:

        """
        Helper function, calls _generate_feature with appropriate parameters based on feature_attributes
        :param feature_attributes: either integer (cardinality) or list of feature attributes
        :param n_samples: number of samples in dataset
        :param ensure_rep: ensures all values are represented at least once in the feature vector
        :param random_values: randomly picked values for vec if true, otherwise values range from [low, cardinality] with by 1
        :param low: lower bound of random feature vector values
        :param high: upper bound of random feature vector values
        :param k: scale constant for normal distribution, default 10, sets width of normal distribution, larger value -> narrower peak
        :return: feature vector
        """

        # feature_cardinality is just an integer, generate feature either with random values or
        # [low, low+cardinality]
        if not isinstance(feature_attributes, (list, np.ndarray)):
            x = self._generate_feature(
                n_samples,
                cardinality=feature_attributes,
                ensure_rep=ensure_rep,
                random_values=random_values,
                low=low,
                high=high,
                k=k,
            )
        # feature_cardinality is a list of [value_domain, value_frequencies]
        else:
            if isinstance(feature_attributes[0], (list, np.ndarray)):
                value_domain, value_frequencies = feature_attributes
                x = self._generate_feature(
                    n_samples,
                    vec=value_domain,
                    ensure_rep=ensure_rep,
                    p=value_frequencies,
                )
            # feature_cardinality is value_domain (list of values for feature)
            else:
                value_domain = feature_attributes
                x = self._generate_feature(
                    n_samples,
                    vec=value_domain,
                    ensure_rep=ensure_rep,
                    k=k,
                )

        return x

    def _generate_feature(
        self,
        size: int,
        vec: list[int] | ArrayLike | None = None,
        cardinality: int = 5,
        ensure_rep: bool = False,
        random_values: bool | None = False,
        low: int | None = 0,
        high: int | None = 1000,
        p: list[float] | np.ndarray | None = None,
        k: int | float = 10,
    ) -> np.ndarray:
        """
        Generates feature vector of length size. Default probability density distribution is approximately normal, centred around a randomly picked value.
        :param vec: list of feature values
        :param cardinality: single value cardinality
        :param size: length of feature vector
        :param ensure_rep: ensures all values are represented at least once in the feature vector
        :param random_values: randomly picked values for vec if true, otherwise values range from [low, cardinality] with by 1
        :param low: lower bound of random feature vector values
        :param high: upper bound of random feature vector values
        :param p: list of probabilities of each value
        :param k: scale constant for normal distribution, default 10, sets width of normal distribution, larger value -> narrower peak
        :return: feature vector x
        """

        if vec is None:
            if random_values:
                vec = range(low, high + 1)
                vec = np.random.choice(vec, size=cardinality, replace=False)
            else:
                vec = np.arange(low, low + cardinality, 1)
        else:
            vec = np.array(vec)

        vec_len = len(vec)
        if p is None:
            v_shift = vec - vec[np.random.randint(len(vec))]
            p = norm.pdf(v_shift, scale=vec_len/k)
        else:
            p = np.array(p)

        p = p / p.sum()

        if ensure_rep and len(vec) < size:
            sampled_values = np.random.choice(vec, size=(size - len(vec)), p=p)
            sampled_values = np.append(sampled_values, vec)
        else:
            sampled_values = np.random.choice(vec, size=size, p=p)

        np.random.shuffle(sampled_values)
        return sampled_values.astype('int32')

    def generate_combinations(
        self,
        X: ArrayLike,
        feature_indices: list[int] | ArrayLike,
        combination_function: Optional = None,
        combination_type: Literal['linear', 'nonlinear'] = 'linear',
    ) -> np.ndarray:
        """
        Generates linear, nonlinear, or custom combinations within feature vectors in given dataset X
        :param X: dataset
        :param feature_indices: indexes of features to be in combination
        :param combination_function: optional custom function for combining feature vectors
        :param combination_type: string flag, either liner or nonlinear, defining combination type
        :return: X with added resultant feature
        """

        selected_features = X[:, feature_indices]

        if combination_function is None:
            if combination_type == 'linear':
                combination_function = lambda x: np.sum(x, axis=1)
            elif combination_type == 'nonlinear':
                combination_function = lambda x: np.sin(np.sum(x, axis=1))
        else:
            combination_type = str(combination_function.__name__)

        combination_result = combination_function(selected_features)

        combination_ix = len(X[0])

        self.dataset_info['combinations'].append({
            'feature_indices': feature_indices,
            'combination_type': combination_type,
            'combination_ix': combination_ix,
        })

        return np.column_stack((X, combination_result))

    def _xor(self, arr: list[int] | ArrayLike) -> np.ndarray:
        """
        Performs bitwise XOR operation on two integer arrays
        :param arr: features to perform XOR operation on
        :return: bitwise XOR result
        """
        arr = np.array(arr)
        arrT = arr.T
        arrT = arrT.astype(int)
        out = np.bitwise_xor(arrT[0], arrT[1])
        if len(arrT) > 2:
            for i in range(2, len(arrT)):
                out = np.bitwise_xor(out, arrT[i])

        return out.T

    def _and(self, arr: list[int] | ArrayLike) -> np.ndarray:
        """
        Performs bitwise AND operation on two integer arrays
        :param arr: features to perform AND operation on
        :return: bitwise AND result
        """
        arr = np.array(arr)
        arrT = arr.T
        arrT = arrT.astype(int)
        out = np.bitwise_and(arrT[0], arrT[1])
        if len(arrT) > 2:
            for i in range(2, len(arrT)):
                out = np.bitwise_and(out, arrT[i])

        return out.T

    def _or(self, arr: list[int] | ArrayLike) -> np.ndarray:
        """
        Performs bitwise OR operation on two integer arrays
        :param arr: features to perform OR operation on
        :return: bitwise OR result
        """
        arr = np.array(arr)
        arrT = arr.T
        arrT = arrT.astype(int)
        out = np.bitwise_or(arrT[0], arrT[1])
        if len(arrT) > 2:
            for i in range(2, len(arrT)):
                out = np.bitwise_or(out, arrT[i])

        return out.T

    def generate_correlated(
        self,
        X: ArrayLike,
        feature_indices: list[int] | ArrayLike,
        r: float = 0.8,
    ) -> np.ndarray:

        """
        Generates correlated features using the given feature indices. Correlation is based on cosine of angle between vectors with mean 0.
        :param X: dataset
        :param feature_indices: indices of features to generate correlated feature to
        :param r: (Pearson) correlation factor
        :return: X with generated correlated  features
        """

        if not isinstance(feature_indices, (list, np.ndarray)):
            feature_indices = np.array([feature_indices])

        if len(feature_indices) > 1:
            correlated_ixs = np.arange(len(X[0]), (len(X[0]) + len(feature_indices)), 1)
        else:
            correlated_ixs = len(X[0])

        selected_features = X[:, feature_indices]
        transposed = np.transpose(selected_features)
        correlated_features = []

        for t in transposed:
            theta = np.arccos(r)
            t_standard = (t - np.mean(t)) / (np.std(t) + 1e-10)

            rand = np.random.normal(0, 1, len(t_standard))
            rand = (rand - np.mean(rand)) / (np.std(rand) + 1e-10)

            M = np.column_stack((t_standard, rand))
            M_centred = (M - np.mean(M, axis=0))

            Id = np.eye(len(t))
            Q = qr(M_centred[:, [0]], mode='economic')[0]
            P = np.dot(Q, Q.T)
            orthogonal_projection = np.dot(Id - P, M_centred[:, 1])
            M_orthogonal = np.column_stack((M_centred[:, 0], orthogonal_projection))

            Y = np.dot(M_orthogonal, np.diag(1 / np.sqrt(np.sum(M_orthogonal ** 2, axis=0))))
            corr = Y[:, 1] + (1 / np.tan(theta)) * Y[:, 0]

            correlated_features.append(corr)

        correlated_features = np.transpose(correlated_features)

        self.dataset_info['correlations'].append({
            'feature_indices': feature_indices,
            'correlated_indices': correlated_ixs,
            'correlation_factor': r,
        })

        return np.column_stack((X, correlated_features))

    def generate_duplicates(
        self,
        X: ArrayLike,
        feature_indices: list[int] | ArrayLike,
    ) -> np.ndarray:
        """
        Generates duplicate features
        :param X: dataset
        :param feature_indices: indices of features to duplicate
        :return: dataset with duplicated features
        """
        if not isinstance(feature_indices, (list, np.ndarray)):
            feature_indices = np.array([feature_indices])

        duplicated_ixs = np.arange(len(X[0]), (len(X[0]) + len(feature_indices) - 1), 1)

        selected_features = X[:, feature_indices]

        self.dataset_info['duplicates'].append({
            'feature_indices': feature_indices,
            'duplicate_indices': duplicated_ixs,
        })

        return np.column_stack((X, selected_features))

    def generate_labels(
        self,
        X: ArrayLike,
        n: int = 2,
        p: float | list[float] | ArrayLike = 0.5,
        k: int | float = 2,
        decision_function: Optional = None,
        class_relation: Literal['linear', 'nonlinear', 'cluster'] = 'linear',
        balance: bool = False,
        random_state: int = 42,
    ):
        """
        Generates labels for dataset X
        :param X: dataset
        :param n: number of class labels
        :param p: class distribution
        :param k: constant
        :param decision_function: optional user-defined decision function
        :param class_relation: string, either 'linear', 'nonlinear', or 'cluster'
        :param balance: boolean, whether to balance clustering class labels
        :param random_state: seed for KMeans clustering, defaults to 42
        :return: array of labels, corresponding to dataset X
        """

        if isinstance(p, (list, np.ndarray)):
            if sum(p) > 1: raise ValueError('sum of values in must be less than 1.0')
            if len(p) > n: raise ValueError('length of p must equal n')
        elif p > 1.0: raise ValueError('p must be less than 1.0')

        n_samples, n_features = X.shape

        if decision_function is None:
            if class_relation == 'linear':
                decision_function = lambda x: np.sum(2 * x + 3, axis=1)
            elif class_relation == 'nonlinear':
                decision_function = lambda x: np.sum(k * np.sin(x) + k * np.cos(x), axis=1)
            elif class_relation == 'cluster':
                decision_function = None
        else:
            class_relation = str(decision_function.__name__)

        y = []
        if decision_function is not None:
            if n > 2:
                if type(p) != list:
                    p = 1 / n
                    percentiles = [p * 100]
                    for i in range(1, n - 1):
                        percentiles.append(percentiles[i - 1] + (p * 100))

                    decision_boundary = decision_function(X)
                    p_points = np.percentile(decision_boundary, percentiles)

                    y = np.zeros_like(decision_boundary, dtype=int)
                    for p_point in p_points:
                        y += (decision_boundary > p_point)
                else:
                    decision_boundary = decision_function(X)
                    percentiles = [x * 100 for x in p]

                    for i in range(1, len(percentiles) - 1):
                        percentiles[i] += percentiles[i - 1]
                    percentiles.insert(0, 0)
                    percentiles.pop()

                    p_points = np.percentile(decision_boundary, percentiles)

                    y = np.zeros_like(decision_boundary, dtype=int)

                    for i in range(1, n):
                        p_point = p_points[i]
                        y += np.where(decision_boundary > p_point, 1, 0)
            else:
                decision_boundary = decision_function(X)
                if isinstance(p, (list, np.ndarray)):
                    p = p[0]
                p_point = np.percentile(decision_boundary, p * 100)
                y = np.where(decision_boundary > p_point, 1, 0)
        else:
            if p == 0.5:
                p = 1.0
            else:
                p = [p, 1 - p]
            y = self._cluster_data(X, n, p=p, balance=balance, random_state=random_state)

        self.dataset_info.update({
            'labels': {
                'class_relation': class_relation,
                'n_class': n,
            },
        })

        return y

    def _cluster_data(
        self,
        X: ArrayLike,
        n: int,
        p: float | list[float] | ArrayLike | None = 1.0,
        balance: bool = False,
        random_state: int = 42,
    ) -> np.ndarray:
        """
        Cluster data using kmeans
        :param X: dataset
        :param n: number of clusters
        :param p: class distribution
        :param balance: balance the clusters according to p
        :random_state: seed for KMeans clustering, defaults to 42
        :return: array of labels, corresponding to dataset X
        """

        kmeans = KMeans(n_clusters=n, random_state=random_state)

        kmeans.fit(X)

        cluster_labels = kmeans.labels_

        # Fully balanced clusters
        if not isinstance(p, (list, np.ndarray)):
            samples_per_cluster = [len(X) // n] * n
        else:
            samples = len(X)
            samples_per_cluster = []
            if not isinstance(p, (list, np.ndarray)):
                samples_per_cluster.append(int(samples * p) // n)
                samples_per_cluster.append(int(samples * (1 - p)) // n)
            else:
                if len(p) == n:
                    for val in p:
                        samples_per_cluster.append(int(samples * val))
                else:
                    raise Exception('Length of balance parameter must equal number of clusters.')

        # Adjust cluster sizes
        if balance:
            adjustments = []
            overflow_samples = []
            overflow_indices = []
            for i in range(n):
                cluster_size = np.sum(cluster_labels == i)

                adjustment = samples_per_cluster[i] - cluster_size
                adjustments.append(adjustment)

                # Cluster is too large
                if adjustment < 0:
                    centroid = kmeans.cluster_centers_[i]
                    # Indices of samples in dataset
                    dataset_indices = np.where(cluster_labels == i)[0]
                    cluster_samples = np.copy(X[dataset_indices])

                    distances = np.linalg.norm(
                        cluster_samples - centroid,
                        axis=1,
                    )  # Distances of cluster samples to cluster centroid
                    cluster_sample_indices = np.argsort(distances)
                    dataset_indices_sorted = dataset_indices[
                        cluster_sample_indices
                    ]  # Indices of samples sorted by sample distance to cluster centroid

                    overflow_sample_indices = cluster_sample_indices[samples_per_cluster[i]:]  # Overflow samples
                    dataset_indices_sorted = dataset_indices_sorted[
                                             samples_per_cluster[i]:
                    ]  # Dataset indices of overflow samples

                    for i in range(len(overflow_sample_indices)):
                        overflow_samples.append(cluster_samples[overflow_sample_indices[i]])
                        overflow_indices.append(dataset_indices_sorted[i])

            overflow_samples = np.array(overflow_samples)
            overflow_indices = np.array(overflow_indices)

            # Making adjustments
            for i in range(n):

                if adjustments[i] > 0:
                    centroid = kmeans.cluster_centers_[i]
                    distances = np.linalg.norm(overflow_samples - centroid, axis=1)

                    closest_sample_indices = np.argsort(distances)

                    overflow_indices_sorted = overflow_indices[closest_sample_indices]

                    sample_indices_slice = closest_sample_indices[:adjustments[i]]
                    overflow_indices_slice = overflow_indices_sorted[:adjustments[i]]

                    cluster_labels[overflow_indices_slice] = i

                    overflow_samples = np.delete(overflow_samples, sample_indices_slice, axis=0)
                    overflow_indices = np.delete(overflow_indices, sample_indices_slice, axis=0)

        return np.array(cluster_labels)

    def generate_noise(
        self,
        X: ArrayLike,
        y: list[int] | ArrayLike,
        p: float = 0.2,
        type: Literal['categorical', 'missing'] = 'categorical',
        missing_val: str | int | float = float('-inf'),
    ) -> np.ndarray:

        """
        Simulates noise on given dataset X
        :param X: dataset to apply noise to
        :param y: required target labels for categorical noise generation
        :param p: amount of noise to apply. Defaults to 0.2
        :param type: type of noise to apply, either categorical or missing
        :param missing_val: value to simulate missing values. Defaults to float('-inf')
        :return: X with noise applied
        """

        self.dataset_info['noise'].append({
            'type': type,
            'amount': p,
        })

        if type == 'categorical':
            label_values, label_count = np.unique(y, return_counts=True)
            n_labels = len(label_values)

            inds = y.argsort()
            y_sort = y[inds]
            X_sort = X[inds]

            Xs_T = X_sort.T
            n = Xs_T.shape[1]
            n_flip = int(n * p)

            for feature in Xs_T:
                unique_per_label = {}

                for i in range(n_labels):
                    if i == 0:
                        unique = np.unique(feature[:label_count[i]])
                        unique_per_label[label_values[i]] = set(unique)
                    else:
                        unique = np.unique(feature[label_count[i - 1]:label_count[i - 1] + label_count[i] - 1])
                        unique_per_label[label_values[i]] = set(unique)

                ixs = np.random.choice(n, n_flip, replace=False)

                for ix in ixs:
                    current_label = y_sort[ix]
                    possible_labels = np.where(label_values != current_label)[0]

                    # find all unique values from labels != current label
                    values = set()
                    for key in possible_labels:
                        values = values.union(unique_per_label[key])

                    # remove any overlapping values, ensuring replacement values are unique & from a target label !=
                    # current label
                    for val in unique_per_label[current_label] & values:
                        values.remove(val)

                    if len(values) > 0:
                        val = np.random.choice(list(values))

                    else:
                        key = possible_labels[np.random.randint(len(possible_labels))]
                        values = unique_per_label[key]
                        val = np.random.choice(list(values))

                    feature[ix] = val

            rev_ind = inds.argsort()
            X_noise = Xs_T.T
            X_noise = X_noise[rev_ind]

            return X_noise

        elif type == 'missing':
            X_noise = np.copy(X)
            Xn_T = X_noise.T
            n = Xn_T.shape[1]
            n_missing = int(n * p)
            #print("n to delete:", n_missing)

            for feature in Xn_T:
                ixs = np.random.choice(n, n_missing, replace=False)

                for ix in ixs:
                    feature[ix] = missing_val

            return Xn_T.T

        else:
            raise ValueError(f'Type {type} not supported')

    def downsample_dataset(
        self,
        X: ArrayLike,
        y: list[int] | ArrayLike,
        n: int | None = None,
        seed: int = 42,
        reshuffle: bool = False,
    ) -> tuple[np.ndarray, np.ndarray]:

        """
        Downsamples dataset X according to N or the number of samples in minority class, resulting in a balanced dataset.
        :param X: Dataset to downsample
        :param y: Labels corresponding to X
        :param n: Optional number of samples per class to downsample to
        :param seed: Seed for random state of resample function
        :param reshuffle: Reshuffle the dataset after downsampling
        :return: Balanced X and y after downsampling
        """

        original_shape = X.shape

        values, counts = np.unique(y, return_counts=True)
        if n is None:
            n = min(counts)

        if n > min(counts):
            raise ValueError('N must be equal to or less than the number of samples in minority class')

        X_arrays_list = []
        y_downsampled = []
        for label in values:
            X_label = [X[i] for i in range(len(y)) if y[i] == label]
            X_label_downsample = resample(
                X_label,
                replace=True,
                n_samples=n,
                random_state=seed,
            )
            X_arrays_list.append(X_label_downsample)
            ys = [label] * n
            y_downsampled = np.concatenate((y_downsampled, ys), axis=0)

        X_downsampled = np.concatenate(X_arrays_list, axis=0)

        if reshuffle:
            indices = np.arange(len(X_downsampled))
            np.random.shuffle(indices)
            X_downsampled = X_downsampled[indices]
            y_downsampled = y_downsampled[indices]

        downsampled_shape = X_downsampled.shape

        self.dataset_info.update({
            'downsampling': {
                'original_shape': original_shape,
                'downsampled_shape': downsampled_shape,
            },
        })

        return X_downsampled, y_downsampled

    def print_dataset(
        self,
        X: ArrayLike,
        y: ArrayLike,
    ):
        """
        Prints given dataset
        :param X: dataset
        :param y: labels
        :return:
        """

        n_samples, n_features = X.shape
        n = 0
        for arr in X:
            print('[', end='')
            for i in range(n_features):
                if i == n_features - 1:
                    print(arr[i], end='')
                else:
                    print(arr[i], end=', ')
            print(f'], Label: {y[n]}')
            n += 1

    # TODO: Logging function
